/** \file user.c

	User-specific file (in this case VisualBoyAdvance).
	Basically process callbacks from the parser.
*/
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <process.h>
extern "C" {
	#include "parser.h"
	#include "lang.h"
	#include "user.h"
}

extern void systemScreenMessage(const char *);

// Global variables
int ColorIt_enabled;
unsigned char ColorIt_tilePalette[384], ColorIt_tileIncluded[384 / 8];
unsigned short ColorIt_tileCustom[384];
unsigned char ColorIt_tileData[384 * 16];
unsigned short ColorIt_palette[NB_PALETTES * 4];
static int showCrc = 0;

void extractPath(char *dest, const char *fileName) {
	int i;
	strcpy(dest, fileName);
	for (i = strlen(dest) - 1; i >= 0; i--) {
		if (dest[i] == '\\' || dest[i] == '/') {
			dest[i + 1] = '\0';
			return;
		}
	}
}

static void init() {
	// Rinitialisation des paramtres
	const unsigned short defaultCol[4] =
		{0x7fff, 0x56b5, 0x2d6b, 0x0000};
	unsigned i;
	memset(ColorIt_tilePalette, 0, sizeof(ColorIt_tilePalette));
	for (i = 0; i < 384; i++)
		ColorIt_tileCustom[i] = i;
	memcpy(ColorIt_palette, defaultCol, sizeof(defaultCol));
}

int ColorIt_init(const char *fullRomPath, const unsigned char *romData) {
	char name[128], fullName[1024];
	unsigned i;

	showCrc = 0;
	ColorIt_systemInit();
	init();

	// rom_directory/cart_name.pal.ini
	memcpy(name, romData + 0x134, 15);
	// Only allowed chars in the cart name
	for (i = 0; name[i]; i++) {
        if (name[i] >= 'A' && name[i] <= 'Z' ||
            name[i] >= 'a' && name[i] <= 'z' ||
            name[i] == ' ' || name[i] == '-' || name[i] == '.' ||
            name[i] >= '0' && name[i] <= '9' ||
			name[i] == 0) {}
		else
			name[i] = '_';
	}
	name[15] = 0;
	strcat(name, ".pal.ini");
	extractPath(fullName, fullRomPath);
	strcat(fullName, name);

	// By default no tile is included to the Crc
	memset(ColorIt_tileIncluded, 0, sizeof(ColorIt_tileIncluded));
	if (ColorIt_enabled = ColorIt_execScript(fullName, "init"))
		startCheckThread(name, fullName);
	return ColorIt_enabled;
}

void ColorIt_execCrc(unsigned crc32) {
	char buf[256];
	sprintf(buf, "[%08x]", crc32);
	init();
	if (!ColorIt_execScript(NULL, buf))			// Si la section n'existe pas
		ColorIt_execScript(NULL, "[default]");	// -> [default]
	// Display the CRC
	if (showCrc) {
		sprintf(buf, "New VRAM CRC: %08x", crc32);
		systemScreenMessage(buf);
	}
}

void ColorIt_ShowError(int line, int col, const char *msg) {
	char buf[1024];
	sprintf(buf, "At line %i, col %i: %s\n", line, col, msg);
	MessageBox(NULL, buf, "Script error", 0);
}

void ColorIt_messageBox(const char *msg) {
	MessageBox(NULL, msg, "Script message", 0);
}

void ColorIt_autoShowVramCrc(int value) {
	showCrc = value;
}

void ColorIt_addTileCrc(unsigned tileStart, unsigned tileEnd) {
	unsigned i;
	for (i = tileStart; i <= tileEnd; i++)
		ColorIt_tileIncluded[i >> 3] |= 1 << (i & 7);
}

void ColorIt_setPalette(unsigned palNo, unsigned long *colors) {
	unsigned short *pal = ColorIt_palette + palNo * 4;
	unsigned i;
	for (i = 0; i < 4; i++)
		// Convert 32 to 15 bits
		pal[i] =
			(colors[i] >> 3 & 0x1f) |
			(colors[i] >> 6 & 0x3e0) |
			(colors[i] >> 9 & 0x7c00);
}

void ColorIt_addTileRule(unsigned tileStart, unsigned tileEnd, unsigned palNo) {
	unsigned i;
	for (i = tileStart; i <= tileEnd; i++)
		ColorIt_tilePalette[i] = palNo;
}

void ColorIt_setTilesetData(unsigned ctNo, unsigned char *data) {
	memcpy(ColorIt_tileData + ctNo * 16, data, 16);
}

void ColorIt_setTile(unsigned tileStart, unsigned tileEnd, unsigned ctNo) {
	unsigned i;
	for (i = tileStart; i <= tileEnd; i++)
		ColorIt_tileCustom[i] = 512 + ctNo++;
}
